from visual import *
from text import *
from string import find
from time import clock

# A graph package for plotting a curve, with labeled axes and autoscaling
# Ultimately this capability should be part of the Visual library
# Bruce Sherwood, Carnegie Mellon University, begun April 2000

# minmax[xaxis][negaxis], minmax[yaxis][negaxis] refer to absolute values of minimum negative values;
# minmax[xaxis][posaxis], minmax[yaxis][posaxis] to maximum positive values.
# This graph package forces the origin (0,0) to be on screen, at the axes intersection.

grey = (0.7,0.7,0.7) # color of axes
tmajor = 20. # length of major tick marks in pixels
tminor = 10. # length of minor tick marks in pixels
hchar = 18. # height of characters in pixels
minorticks = 5. # number of minor tick intervals between major ticks
maxmajorticks = 3 # max number of major ticks (not including 0)
maxminorticks = (maxmajorticks+1)*minorticks # max number of minor ticks (4 between major ticks)
lastgdisplay = None # the most recently created gdisplay
gdotsize = 6.0 # width (diameter if gdot were round) of gdot in pixels
zdepth = 1.0 # z range of scene
dz = 0.02 # offset for plots relative to axes and labels
xaxis = 0
yaxis = 1
negaxis = 0
posaxis = 1
reset = [[0, 0], [0, 0]]
graphaxes = 0 # set to 1 when first axes drawn

def labelnum(x): # determine what labels to show, in what format
    if x >= 0.: sign = 1.
    else: sign = -1.
    mantissa, exponent = modf(log10(abs(x)))
    number = 10.**mantissa
    if number < 1.:
        number = 10.*number
        exponent = exponent-1.
    if number >= 7.5: 
        number = 7.5
        marks = [2.5, 5.0, 7.5]
        extra = 1
    elif number >= 5.: 
        number = 5.
        marks = [2.5, 5.0]
        extra = 1
    elif number >= 4.:
        number = 4.
        marks = [2.0, 4.0]
        extra = 0
    elif number >= 3.:
        number = 3.
        marks = [1.0, 2.0, 3.0]
        extra = 0
    elif number >= 2.:
        number = 3.
        marks = [1.0, 2.0]
        extra = 0
    elif number >= 1.5:
        number = 1.5
        marks = [0.5, 1.0, 1.5]
        extra = 1
    else: 
        number = 1.
        marks = [0.5, 1.0]
        extra = 1
    if exponent > 0:
        digits = 0
    else:
        digits = int(-exponent)+extra
    if digits < 3. and exponent <= 3.:
        format = '%0.'+('%s' % digits)+'f'
    else:
        format = '%0.1E'
    return (sign*(array(marks)*10.**exponent)).tolist(), format

def cleaneformat(string): # convert 2.5E-006 to 2.5E-6; 2.5E+006 to 2.5E6
    index = find(string, 'E')
    if index == -1: return string # not E format
    index = index+1
    if string[index] == '-':
        index = index+1
    elif string[index] == '+':
        string = string[:index]+string[index+1:]
    while string[index] == '0':
        string = string[:index]+string[index+1:]
    return string

class gdisplay:
    def __init__(self, x=0, y=0, width=800, height=400, xtitle=None, ytitle=None,
                 xmax=None, xmin=None, ymax=None, ymin=None):
        global lastgdisplay, graphaxes
        lastgdisplay = self
        self.width = float(width)
        self.height = float(height)
        self.display = display(title='Graph', x=x, y=y,
                               width=self.width, height=self.height,
                               uniform=0, fov=0.001)
        self.display.autoscale = 0
        self.autoscale = [1, 1]
        self.title = [[xtitle, 0], [ytitle, 0]] # titles not displayed yet
        # For all axis-related quantities: [x axis 0 or y axis 1][neg axis 0 or pos axis 1]
        self.axis = [[None, None], [None, None]]
        self.zero = [None, None]
        self.lastlabel = [[0., 0.], [0., 0.]]
        self.format = [None, None]
        self.majormarks = [[None, None], [None, None]]
        self.minmax = [[0., 0.], [0., 0.]] # [x or y][negative 0 or positive 1]
        if xmax <> None:
            if xmax > 0:
                self.minmax[xaxis][posaxis] = xmax
                self.autoscale[xaxis] = 0
        if xmin <> None:
            if xmin < 0:
                self.minmax[xaxis][negaxis] = xmin
                self.autoscale[xaxis] = 0
        if ymax <> None:
            if ymax > 0:
                self.minmax[yaxis][posaxis] = ymax
                self.autoscale[yaxis] = 0
        if ymin <> None:
            if ymin < 0:
                self.minmax[yaxis][negaxis] = ymin
                self.autoscale[yaxis] = 0

        self.minorticks = [ [ [], [] ], [ [],[] ] ] # all the minor ticks we'll ever use
        for axis in range(2):
            for axissign in range(2):
                for nn in range(maxminorticks):
                    self.minorticks[axis][axissign].append(curve(pos=[(0,0,0),(0,1,0)], color=grey, visible=0))

        self.majorticks = [ [ [], [] ], [ [],[] ] ] # all the major ticks we'll ever use
        for axis in range(2):
            for axissign in range(2):
                for nn in range(maxmajorticks):
                    self.majorticks[axis][axissign].append(curve(pos=[(0,0,0),(0,1,0)], color=grey, visible=0))
                    self.majorticks[axis][axissign][nn].label = None

        if self.autoscale == [0, 0]:
            reset = [[1, 1,], [1, 1]]
            self.setrange()
            for axis in range(2):
                for axissign in range(2):
                    self.axisdisplay(axis, axissign)
            graphaxes = 1

    def setxyparams(self):
        global x0, y0, tminorx, tminory, tmajorx, tmajory, charheight, charwidth
        if self.minmax[xaxis][posaxis] >= 0 and self.minmax[xaxis][negaxis] >= 0:
            x0 = 0.
        else:
            x0 = self.minmax[xaxis][negaxis]
        if self.minmax[yaxis][posaxis] >= 0 and self.minmax[yaxis][negaxis] >= 0:
            y0 = 0.
        else:
            y0 = self.minmax[yaxis][negaxis]
        
        tminorx = tminor*(self.minmax[yaxis][posaxis]+self.minmax[yaxis][negaxis])/self.height
        tminory = tminor*(self.minmax[xaxis][posaxis]+self.minmax[xaxis][negaxis])/self.width
        tmajorx = tmajor*(self.minmax[yaxis][posaxis]+self.minmax[yaxis][negaxis])/self.height
        tmajory = tmajor*(self.minmax[xaxis][posaxis]+self.minmax[xaxis][negaxis])/self.width

        charheight = hchar*(self.minmax[yaxis][posaxis]+self.minmax[yaxis][negaxis])/self.height
        charwidth = charheight*((self.minmax[xaxis][posaxis]+self.minmax[xaxis][negaxis])/
                (self.minmax[yaxis][posaxis]+self.minmax[yaxis][negaxis]))*(self.height/self.width)
        
    def axisdisplay(self, axis, axissign):
        if self.minmax[axis][axissign] == 0.: return
        self.setxyparams()
        if axissign == posaxis: sign = 1.
        else: sign = -1.
        if self.axis[axis][axissign] == None: # new; no axis displayed up till now
            if axis == xaxis:
                axispos = [(0,y0,0), (sign*self.minmax[axis][axissign],y0,0)]
                titlepos=(self.minmax[axis][posaxis]+0.03*(self.minmax[axis][posaxis]+self.minmax[axis][negaxis]),y0,0)
            else:
                axispos = [(x0,0,0), (x0,sign*self.minmax[axis][axissign],0)]            
                titlepos=(x0,self.minmax[axis][axissign]+0.03*(self.minmax[axis][posaxis]+self.minmax[axis][negaxis])) 
            self.axis[axis][axissign] = curve(pos=axispos, color=grey, display=self.display)
            if (self.title[axis][0] <> None) and (not self.title[axis][1]):
                self.title[axis][0] = text(display=self.display,
                       pos=titlepos, string=self.title[axis][0],
                       color=color.white, height=charheight, width=charwidth)
                self.title[axis][1] = 1

            if self.minmax[axis][posaxis] >= self.minmax[axis][negaxis]:
                newmajormarks, format = labelnum(self.minmax[axis][posaxis])
            else:
                newmajormarks, format = labelnum(self.minmax[axis][negaxis])
            if self.format[axis] == None:
                self.format[axis] = format
                
            if self.zero[axis] == None:
                if axis == xaxis:
                    self.zero[axis] = self.showtext(axis, 0, (0,y0-tminorx-charheight,0))
                elif self.minmax[xaxis][negaxis] == 0:
                    self.zero[axis] = self.showtext(axis, 0, (x0-charwidth/4.,0-charheight/2.,0))

            d1 = newmajormarks[0]
            d2 = d1/minorticks
            nminor = 0
            nmajor = 0
            marks = []
            for x1 in newmajormarks:
                if x1 > self.minmax[axis][axissign]: break # newmajormarks can refer to opposite half-axis
                marks.append(x1)
                if axis == xaxis:
                    position = [(sign*x1,y0,0),(sign*x1,y0+tmajorx,0)]
                else:
                    position = [(x0,sign*x1,0),(x0+tmajory,sign*x1,0)]
                self.majorticks[axis][axissign][nmajor].pos = position
                self.majorticks[axis][axissign][nmajor].visible = 1
                if axis == xaxis:
                    position = (sign*x1,y0-tminorx-charheight,0)
                else:
                    position = (x0-charwidth/4.,sign*x1-charheight/2.,0)
                self.majorticks[axis][axissign][nmajor].label = self.showtext(axis, sign*x1, position)
                nmajor = nmajor+1

            nminor = 0
            for x2 in arange(0.+d2, self.minmax[axis][axissign]+1.01*d2, d2):
                if x2 > self.minmax[axis][axissign]+0.01*d2: break
                if x2 % d1 < d2/2.: continue # don't put minor tick where there is a major one
                if axis == xaxis:
                    self.minorticks[axis][axissign][nminor].pos = [(sign*x2,y0,0),(sign*x2,y0+tminorx,0)]
                else:
                    self.minorticks[axis][axissign][nminor].pos = [(x0,sign*x2,0),(x0+tminory,sign*x2,0)]
                self.minorticks[axis][axissign][nminor].visible = 1
                nminor = nminor+1
                    
            if marks <> []:
                self.majormarks[axis][axissign] = marks
                self.lastlabel[axis][axissign] = self.majormarks[axis][axissign][-1]
            else:
                self.lastlabel[axis][axissign] = 0

        else:
            if reset[axis][axissign]:
                if axis == xaxis:
                    self.axis[axis][axissign].pos = [(0,y0,0), (sign*self.minmax[axis][axissign],y0,0)]
                else:
                    self.axis[axis][axissign].pos = [(x0,0,0), (x0,sign*self.minmax[axis][axissign],0)]
                
            if self.title[axis][1] and axissign == posaxis:
                if axis == xaxis:
                    titlepos=(self.minmax[axis][posaxis]+0.03*(self.minmax[axis][posaxis]+self.minmax[axis][negaxis]),y0,0)
                else:
                    titlepos=(x0,self.minmax[axis][axissign]+0.03*(self.minmax[axis][posaxis]+self.minmax[axis][negaxis])) 
                self.title[axis][0].reshape(pos=titlepos, height=charheight, width=charwidth)

            obj = self.zero[axis] # adjust 0 label
            if obj <> None:
                pos = obj.frame.pos
                if axis == xaxis:
                    pos.y = y0-tminorx-charheight
                else:
                    pos.x = x0-charwidth/4.
                    pos.y = pos.y+obj.height/2.-charheight/2.
                if axis == yaxis and self.minmax[xaxis][negaxis] <> 0:
                    obj.makeinvisible()
                    self.zero[axis] = None
                else:
                    obj.reshape(pos=pos, height=charheight, width=charwidth)

            if self.minmax[axis][posaxis] >= self.minmax[axis][negaxis]:
                newmajormarks, format = labelnum(self.minmax[axis][posaxis])
            else:
                newmajormarks, format = labelnum(self.minmax[axis][negaxis])
            while newmajormarks[-1] > self.minmax[axis][axissign]:
                if len(newmajormarks) == 1: break
                del newmajormarks[-1]
            d1 = newmajormarks[0]
            d2 = d1/minorticks
                
            newformat = (format <> self.format[axis])
            self.format[axis] = format
            needminor = (self.minmax[axis][axissign] >= self.lastlabel[axis][axissign]+d2)
            needmajor = (self.majormarks[axis][axissign] == None) or ((newmajormarks[-1] <> self.majormarks[axis][axissign][-1]) or newformat)

            # create or rescale all minor and major tick marks if necessary
            if needmajor or reset[not axis][axissign]:
                nmajor = 0
                for x1 in newmajormarks:
                    if x1 > self.minmax[axis][axissign]: break
                    if axis == xaxis:
                        self.majorticks[axis][axissign][nmajor].pos = [(sign*x1,y0,0),(sign*x1,y0+tmajorx,0)]
                    else:
                        self.majorticks[axis][axissign][nmajor].pos = [(x0,sign*x1,0),(x0+tmajory,sign*x1,0)]
                    self.majorticks[axis][axissign][nmajor].visible = 1
                    nmajor = nmajor+1
                while nmajor < maxmajorticks:
                    self.majorticks[axis][axissign][nmajor].visible = 0
                    nmajor = nmajor+1
                    
            if needmajor or needminor or reset[not axis][axissign]:
                nminor = 0
                for x2 in arange(0.+d2, self.minmax[axis][axissign]+1.01*d2, d2):
                    if x2 > self.minmax[axis][axissign]+0.01*d2: break
                    if x2 % d1 < d2/2.: continue # don't put minor tick where there is a major one
                    if needmajor or reset[not axis][axissign] or x2 > self.lastlabel[axis][axissign]-0.01*d2:
                        if axis == xaxis:
                            self.minorticks[axis][axissign][nminor].pos = [(sign*x2,y0,0),(sign*x2,y0+tminorx,0)]
                        else:
                            self.minorticks[axis][axissign][nminor].pos = [(x0,sign*x2,0),(x0+tminory,sign*x2,0)]
                        self.minorticks[axis][axissign][nminor].visible = 1
                    nminor = nminor+1
                while nminor < maxminorticks:
                    self.minorticks[axis][axissign][nminor].visible = 0
                    nminor = nminor+1

            self.lastlabel[axis][axissign] = d2*int(self.minmax[axis][axissign]/d2)

            if needmajor: # need new labels
                if self.majormarks[axis][axissign] == None:
                    oldd1 = 0
                else:
                    oldd1 = self.majormarks[axis][axissign][0]
                marks = []
                for nmajor in range(maxmajorticks):
                    x1 = (nmajor+1)*d1
                    if x1 > self.minmax[axis][axissign]:
                        if self.majorticks[axis][axissign][nmajor].label <> None: # get rid of unused labels
                            self.majorticks[axis][axissign][nmajor].label.makeinvisible()
                            self.majorticks[axis][axissign][nmajor].label = None
                        continue
                    marks.append(x1)
                    if (self.majormarks[axis][axissign] <> None) and (x1 in self.majormarks[axis][axissign]): # reuse a label if it already exists
                        if oldd1 <> d1:
                            index = self.majormarks[axis][axissign].index(x1)
                            obj = self.majorticks[axis][axissign][index].label
                            self.majorticks[axis][axissign][index].label = None
                            self.majorticks[axis][axissign][nmajor].label.makeinvisible()
                        else:
                            obj = self.majorticks[axis][axissign][nmajor].label
                        pos = obj.frame.pos
                        if axis == xaxis:
                            pos.y = y0-tminorx-charheight
                        else:
                            pos.x = x0-charwidth/4.
                            pos.y = pos.y+obj.height/2.-charheight/2.
                        obj.reshape(pos=pos, height=charheight, width=charwidth)
                        self.majorticks[axis][axissign][nmajor].label = obj
                    else: # else create a new label
                        obj = self.majorticks[axis][axissign][nmajor].label
                        if obj <> None:
                            obj.makeinvisible()
                        if axis == xaxis:
                            self.majorticks[axis][axissign][nmajor].label = self.showtext(axis, sign*x1, (sign*x1,y0-tminorx-charheight,0))
                        else:
                            self.majorticks[axis][axissign][nmajor].label = self.showtext(axis, sign*x1, (x0-charwidth/4.,sign*x1-charheight/2.,0))

                if marks <> []:
                    self.majormarks[axis][axissign] = marks
                        
            else: # adjust existing labels
                for nmajor in range(maxmajorticks):
                    obj = self.majorticks[axis][axissign][nmajor].label
                    if obj <> None:
                        pos = obj.frame.pos
                        if axis == xaxis:
                            pos.y = y0-tminorx-charheight
                        else:
                            pos.x = x0-charwidth/4.
                            pos.y = pos.y+obj.height/2.-charheight/2.
                        obj.reshape(pos=pos, height=charheight, width=charwidth)
                        
    def showtext(self, axis, val, pos):
        if val == 0.:
            string = '0'
        else:
            string = cleaneformat(self.format[axis] % val)
        if axis == xaxis:
            format = 'center'
        else:
            format = 'right'
        return text(display=self.display,
                    pos=pos, string=string,
                    color=color.white, height=charheight,
                    width=charwidth, justify=format)
            
    def setrange(self):
        self.display.range = (0.65*(self.minmax[xaxis][posaxis]+self.minmax[xaxis][negaxis]),0.65*(self.minmax[yaxis][posaxis]+self.minmax[yaxis][negaxis]),zdepth)
        self.display.center = ((self.minmax[xaxis][posaxis]-self.minmax[xaxis][negaxis])/2.0,(self.minmax[yaxis][posaxis]-self.minmax[yaxis][negaxis])/2.0,zdepth)
           
    def resize(self, x=None, y=None, color=None, gcurve=None):
        global reset, graphaxes
        reset = [[0, 0], [0, 0]]
        redo = 0
        if self.autoscale[xaxis]:
            if x > self.minmax[xaxis][posaxis]:
                self.minmax[xaxis][posaxis] = x
                reset[xaxis][posaxis] = redo = 1
            elif -x > self.minmax[xaxis][negaxis]:
                self.minmax[xaxis][negaxis] = -x
                reset[xaxis][negaxis] = redo = 1
        if self.autoscale[yaxis]:
            if y > self.minmax[yaxis][posaxis]:
                self.minmax[yaxis][posaxis] = y
                reset[yaxis][posaxis] = redo = 1
            elif -y > self.minmax[yaxis][negaxis]:
                self.minmax[yaxis][negaxis] = -y
                reset[yaxis][negaxis] = redo = 1
                
        if not graphaxes:
            for axis in range(2):
                for axissign in range(2):
                    if self.axis[axis][axissign] == None:
                        if (self.minmax[axis][posaxis] + self.minmax[axis][negaxis]) == 0.: return
                    reset[axis][posaxis] = reset[axis][negaxis] = redo = 1
            graphaxes = 1
             
        if redo:
            self.setrange()
            for axis in range(2):
                for axissign in range(2):
                    self.axisdisplay(axis,axissign)
##            self.display.mouse.getclick()

def getgdisplay():
    return gdisplay()

class gcurve:
    def __init__(self, gdisplay=None, color=None):
        if gdisplay == None:
            if lastgdisplay == None:
                gdisplay = getgdisplay()
            else:
                gdisplay = lastgdisplay
        self.gdisplay = gdisplay
        self.color = color
        self.gcurve = curve(display=gdisplay.display, color=color)

    def plot(self, x=None, y=None, color=None):
        if color == None:
            color = self.color
        self.gdisplay.resize(x, y)
        self.gcurve.append(x=x, y=y, color=color)
                    
class gdot:
    def __init__(self, gdisplay=None, color=None):
        if gdisplay == None:
            if lastgdisplay == None:
                gdisplay = getgdisplay()
            else:
                gdisplay = lastgdisplay
        self.gdisplay = gdisplay
        self.color = color
        self.gdot = []
        self.save()

    def save(self):
        self.lastxmax = self.gdisplay.minmax[xaxis][posaxis]
        self.lastxmin = self.gdisplay.minmax[xaxis][negaxis]
        self.lastymax = self.gdisplay.minmax[yaxis][posaxis]
        self.lastymin = self.gdisplay.minmax[yaxis][negaxis]  

    def plot(self, x=None, y=None, color=None):
        # should give gdisplay gdot reference and have it drive resizing of dots when graph resizes............
        if color == None:
            color = self.color
        self.gdisplay.resize(x, y)
        rx = gdotsize*(self.gdisplay.minmax[xaxis][posaxis]+self.gdisplay.minmax[xaxis][negaxis])/self.gdisplay.width
        ry = gdotsize*(self.gdisplay.minmax[yaxis][posaxis]+self.gdisplay.minmax[yaxis][negaxis])/self.gdisplay.height
        if rx == 0: rx = 1. # avoid problems with initial display
        if ry == 0: ry = 1.
        if (self.gdisplay.minmax[xaxis][posaxis] <> self.lastxmax) or (self.gdisplay.minmax[xaxis][negaxis] <> self.lastxmin)\
           or (self.gdisplay.minmax[yaxis][posaxis] <> self.lastymax) or (self.gdisplay.minmax[yaxis][negaxis] <> self.lastymin):
            for dot in self.gdot:
                x1 = dot[0]
                y1 = dot[1]
                dot[2].size = (rx, ry, dz/4.)
        self.gdot.append([x, y, box(pos=(x,y,-dz/2.), size=(rx, ry, dz/4.), color=color)])
        self.save()

class gvbar:
    def __init__(self, delta=1, gdisplay=None, color=None):
        if gdisplay == None:
            if lastgdisplay == None:
                gdisplay = getgdisplay()
            else:
                gdisplay = lastgdisplay
        self.gdisplay = gdisplay
        self.color = color
        self.delta = delta
        self.vbars = []

    def plot(self, x=None, y=None, color=None):
        if color == None:
            color = self.color
        self.gdisplay.resize(x, y)
        self.vbars.append(box(pos=(x,y/2.,-dz), size=(self.delta,y,dz), color=color))
        
class ghbar:
    def __init__(self, delta=1, gdisplay=None, color=None):
        if gdisplay == None:
            if lastgdisplay == None:
                gdisplay = getgdisplay()
            else:
                gdisplay = lastgdisplay
        self.gdisplay = gdisplay
        self.color = color
        self.delta = delta
        self.hbars = []

    def plot(self, x=None, y=None, color=None):
        if color == None:
            color = self.color
        self.gdisplay.resize(x, y)
        self.hbars.append(box(pos=(x/2.,y,-dz), size=(x,self.delta,dz), color=color))

class ghistogram:
    def __init__(self, bins=None, accumulate=0, average=0, gdisplay=None, color=None):
        if gdisplay == None:
            if lastgdisplay == None:
                gdisplay = getgdisplay()
            else:
                gdisplay = lastgdisplay
        self.gdisplay = gdisplay
        self.bins = bins
        self.nhist = 0 # number of calls to plot routine
        self.accumulate = accumulate # add successive data sets
        self.average = average # display accumulated histogram divided by self.nhist
        self.color = color
        self.histaccum = zeros(len(bins))
        self.delta = (bins[1]-bins[0])
        self.vbars = []

    def plot(self, data=None, accumulate=None, average=None, color=None):
        if color == None:
            color = self.color
        if accumulate <> None:
            self.accumulate = accumulate
        if average <> None:
            self.average = average
        n = searchsorted(sort(data), self.bins)
        # need to resize graph.............................not done yet
        n = concatenate([n, [len(data)]])
        histo = n[1:]-n[:-1]
        if self.accumulate:
            self.histaccum = self.histaccum+histo
        else:
            self.histaccum = histo
        self.nhist = self.nhist+1.
        for nbin in range(len(self.bins)):
            x = (nbin+0.5)*self.delta
            y = self.histaccum[nbin]
            if self.nhist == 1.:
                self.vbars.append(box(display=self.gdisplay.display, pos=(x,y/2.,-dz),
                    size=(0.8*self.delta,0,dz), color=color))
            else:
                if self.accumulate and self.average: 
                    y = y/self.nhist
                self.vbars[nbin].y = y/2.
                self.vbars[nbin].height = y

if __name__ == '__main__':
    # If xmax, xmin, ymax, or ymin specified, the related axis is not autoscaled
    # Can turn off autoscaling with
    #    oscillation.autoscale[0]=0 for x or oscillation.autoscale[1]=0 for y
    oscillation = gdisplay(xtitle='TIME', ytitle='RESPONSE')
    funct1 = gcurve(color=color.cyan)
    funct2 = gvbar(delta=0.5, color=color.red)
    funct3 = gdot(color=color.yellow)

    tt = clock()
    for t in arange(-30, 70, 1):
        funct1.plot(t, 5.0+5.0*cos(-0.2*t)*exp(0.015*t) )
        funct2.plot(t, 2.0+5.0*cos(-0.1*t)*exp(0.015*t) )
        funct3.plot(t, 5.0*cos(-0.03*t)*exp(0.015*t) )
    tt = clock()-tt
    print tt # -32, 70, 1 is 11 seconds on BAS home PC
        
