#include "gldevice.h"
#include "prim.h"

/***************** GLDevice *******************/

GLDevice::GLDevice()
 : active(false),
   manual_scale(1.0),
   initx(-1), inity(-1),
   initw(430), inith(430),
   quitOnClose(0),
   last_wct( tmatrix() )
{
  tmatrix z; 
  z.x_column(0,0,0);
  z.y_column(0,0,0);
  z.z_column(0,0,0);
  z.w_column(0,0,0);
  z.w_row(0,0,0);
  last_wct = z;
}

GLDevice::~GLDevice() {
  hide();
  join();
}

bool GLDevice::show() {
  if (active && mode==DISPLAY) 
    return true;

  mode = SHOW;
  addCallback();
  join();
  if (mode == DISPLAY) {
    addCallback();
    return true;
  } else {
    PySys_WriteStderr("OpenGL initialization failed.\n"
                      "  %s\n", error_message.c_str());
    return false;
  }

  return false;
}

string GLDevice::info() {
  if (closed()) {
    return string("Renderer inactive.\n");
  } else {
    string s;
    s += "OpenGL renderer active.\n";
    s += format("  Vendor: %s\n"
                "  Version: %s\n"
                "  Renderer: %s\n"
                "  Extensions:\n",
                vendor.c_str(), version.c_str(), renderer.c_str());
  
    for(int count=0, p=0; extensions[p] && count<25; count++) {
      int space = extensions.find(' ', p);
      if (space == string::npos) space = extensions.size();
      string token(extensions, p, space-p);
      s += format("%39s\n", token.c_str());
      if (extensions[space]!=' ') break;
      p = space + 1;
    }
    return s;
  }
}

void GLDevice::hide() {
  mode = HIDE;
}

bool GLDevice::closed() {
  return !active;
}

void GLDevice::join() {
  while (active) {
    sleep(0.100);
  }
}

void GLDevice::frame() {
  if (closed()) {
    show();
    return;
  }
  mode = FRAME;
  join();
  if (mode == DISPLAY)
    addCallback();
}

void GLDevice::addCallback() {
  active = true;
  threaded_timer( 0, &GLDevice::callback, this );
}

void GLDevice::setX(int v) {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  initx = v;
}
void GLDevice::setY(int v) {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  inity = v;
}
void GLDevice::setWidth(int v) {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  initw = v;
}
void GLDevice::setHeight(int v) {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  inith = v;
}

int GLDevice::getX() {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  return initx;
}
int GLDevice::getY() {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  return inity;
}
int GLDevice::getWidth() {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  return initw;
}
int GLDevice::getHeight() {
  if (active) throw RuntimeError("Window attributes are not accessible once the window has been created.\n");
  return inith;
}

/******* The following methods are called from a threaded_timer, potentially
         in parallel with calls to the preceding methods. *******************/

void GLDevice::getProjection(tmatrix& wct, Vector& cam) {
  float hfov, vfov;
  float aspect = float(cx.height()) / cx.width();
  if (aspect < 1.0) {
    hfov = display->tanfov;
    vfov = hfov * aspect;
  } else {
    vfov = display->tanfov;
    hfov = vfov / aspect;
  }
  float cotfov = 1.0 / display->tanfov;
  // xxx this calculation works only if autoscale is on: it
  //     does not account for the fact that range might be different
  //     from extent.
  float ext = 1.5 * manual_scale;
  float farclip = cotfov + ext,
        nearclip = cotfov - ext*1.5;
  if (nearclip < 0.01*farclip) nearclip = 0.01*farclip;

  //tmatrix proj, iproj;
  double R = nearclip*hfov,
         T = nearclip*vfov;
  frustum(proj,iproj,-R,R,-T,T,nearclip,farclip);

  if (display->c_uniform) {
    manual.x_column(manual_scale,0,0);
    manual.y_column(0,manual_scale,0);
    manual.z_column(0,0,manual_scale);
  } else if (cx.height() < cx.width()) {
    manual.x_column(manual_scale,0,0);
    manual.y_column(0,manual_scale * cx.height() / cx.width(), 0);
    manual.z_column(0,0,manual_scale * cx.height() / cx.width());
  } else {
    manual.x_column(manual_scale * cx.width() / cx.height(), 0);
    manual.y_column(0,manual_scale,0);
    manual.z_column(0,0,manual_scale);
  }
  manual.w_column();
  manual.w_row();

  // wct = display->model o manual o display->view o proj
  tmatrix model(display->model, manual);
  tmatrix modelview(model, display->view);
  wct = tmatrix( modelview, proj );

  // the camera is at [0 0 0 1] in view space, so we
  //   calculate (model^-1) (manual^-1) (view^-1) [0 0 0 1]T
  Vector iscale(1.0/manual[0][0], 1.0/manual[1][1], 1.0/manual[2][2]);
  cam = display->imodel * ((display->iview * Vector(0,0,0)).scale(iscale));
}

Vector GLDevice::calcMousePos(Vector mpos) {
  Vector p(mpos.x*2-1, 1-mpos.y*2, 0.5);
  p = (iproj * p)/iproj.w(p);
  /* p is in view space.  The camera is at [0,0,0]; the
     center is at c=[0 0 -1/tanfov].  We scale p so that
     p.z == -1/tanfov, without changing its direction. */
  p = p * (-1/display->tanfov) / p.z;
  
  p = display->iview * p;
  p.x /= manual[0][0]; p.y /= manual[1][1]; p.z /= manual[2][2];
  p = display->imodel * p;

  return p;
}

void GLDevice::mouseControl(Vector pos, Vector cam, Vector ray, Object pick, Vector pickpos) {
  Vector delta = cx.getMouseDelta();
  unsigned buttons = cx.getMouseButtons();
  unsigned changed = cx.getMouseButtonsChanged();
  if (buttons) {
    cx.lockMouse();
    clickDelta -= delta.mag();
  } else {
    cx.unlockMouse();
  }

  {
    mouseObject *m = display->mouse.extensionObject();
    mutex::lock L(m->mtx);

    m->position = pos;
    m->pick = pick;
    m->pickpos = pickpos;
    m->cam = cam;
    m->ray = ray;

    if (changed & ~buttons && clickDelta>=0) {
      // Mouse click!

      clickObject *c = new clickObject();
      c->position = m->position;
      c->pick = m->pick;
      c->pickpos = m->pickpos;
      c->cam = cam;
      c->ray = ray;

      m->clicks.push_back(ExtensionObject<clickObject>(FromAPI(c)));
    }

    if (!buttons) clickDelta = 0.03F;
  }

  if (buttons & 2) {
    Vector forward = display->c_forward, up = display->c_up;
    tmatrix m;
    rotation(m, delta.x*-2, up.norm());
    forward = m * forward;

    if (!(buttons&1)) {
      float ya = -delta.y*2;
      float limit = acos( norm_dot(up,forward) );
      if (ya < limit-pi+0.01) ya=limit-pi+0.01;
      if (ya > limit-0.01) ya=limit-0.01;

      rotation(m, ya, forward.cross(up).norm());
      forward = m * forward;
    }
    display->setForward( forward );
  }

  if (buttons & 1) {
    manual_scale *= pow(10, -delta.y);
  }

  oldButtons = buttons;
}

int GLDevice::render() {
  if (mode == HIDE) {
    cx.cleanup();
    return -1;
  }

  { DisplayObject::read_lock D(display->mtx);
    display->updateCache();
  }

  if (mode == SHOW) {
    if( !cx.initWindow(display->c_title.c_str(),initx,inity,initw,inith))
    {
      error_message = cx.lastError();
      return -1;
    }
    cx.makeCurrent();
    vendor = (const char*)glGetString(GL_VENDOR);
    version = (const char*)glGetString(GL_VERSION);
    renderer = (const char*)glGetString(GL_RENDERER);
    extensions = (const char*)glGetString(GL_EXTENSIONS);
    cx.makeNotCurrent();
    mode = DISPLAY;
    return -1;
  }
  
  if (!cx.isOpen()) {
    cx.cleanup();
    if (quitOnClose)
      threaded_exit(0);
    return -1;
  }

  if (display->c_title_changed)
    cx.changeWindow( display->c_title.c_str(), -1, -1, -1, -1 );

  cx.makeCurrent();
  glViewport( 0, 0, cx.width(), cx.height() );
  glClearColor( display->bgcolor.r, display->bgcolor.g, display->bgcolor.b, 1 );
  glClearDepth( 1 );
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable(GL_DEPTH_TEST);
  glDepthFunc( GL_LEQUAL );

  tmatrix wct;
  Vector cam;
  getProjection(wct, cam);
  last_wct = wct;

  Vector mpos = calcMousePos(cx.getMousePos());
  Vector ray = (mpos - cam).norm0();
  double best_intersect = 1e300;
  Object pick;

  Vector min_ext(display->c_center), max_ext(display->c_center);
  rView view(wct, display->lights, tmatrix::identity(), min_ext, max_ext);

  DisplayListIterator p(display->objects());
  for(p++; p; p++) {
    p->updateCache();
    
    DisplayObject* parent = p->getParent();
    if (parent) {
      parent->updateCache();
      tmatrix iframe, frame = parent->getChildTransform();
      iframe.invert_ortho(frame);
      rView local( tmatrix(frame, wct),
                   lighting(view.lights, iframe),
                   frame,
                   view.min_extent,
                   view.max_extent );
      p->glRender( local );

      view.ext_set(local.min_extent, local.max_extent);
    } else
      p->glRender( view );

    if (view.min_extent[0] < min_ext.x) min_ext.x = view.min_extent[0];
    if (view.min_extent[1] < min_ext.y) min_ext.y = view.min_extent[1];
    if (view.min_extent[2] < min_ext.z) min_ext.z = view.min_extent[2];
    if (view.max_extent[0] > max_ext.x) max_ext.x = view.max_extent[0];
    if (view.max_extent[1] > max_ext.y) max_ext.y = view.max_extent[1];
    if (view.max_extent[2] > max_ext.z) max_ext.z = view.max_extent[2];

    double dist = p->rayIntersect(cam, ray);
    if (dist>0.0 && dist < best_intersect) {
      best_intersect = dist;
      pick = p->getObject();
    }
  };

  display->setRegion(min_ext, max_ext);

  //glColor4f(1,1,1,1);
  //glRasterPos2f( -1, -1 );
  //cx.glPrintf(cx.sysFont, "%d", cx.getMouseButtons());

  cx.swapBuffers();
  cx.makeNotCurrent();

  mouseControl(mpos, cam, ray, pick, cam + ray*best_intersect);

  if (mode == FRAME) {
    mode = DISPLAY;
    return -1;
  }

  return 33;
}

bool GLDevice::callback( GLDevice *dev ) {
  if (dev->active) {
    try {
      int next = dev->render();
      if ( next >= 0 ) {
        threaded_timer( next*1e-3, &GLDevice::callback, dev );
      } else {
        dev->active = false;
      }
    } catch (...) {
      dev->active = false;
    }
  }

  return false;  // don't call me again
}
