#include "wgl.h"

using namespace win;
#include <process.h>

#define clamp(x, a, b) (__max(a, __min(x,b) ))

int wglContext::sysFont = -1;

void wglContext::glPrintf(int font, char* format, ...) {
  if (font<0) return;

  va_list a;
  va_start(a,format);
  
  char buffer[1024];
  vsprintf(buffer, format, a);
  int len = strlen(buffer);
  glListBase(font);
  glCallLists( len, GL_UNSIGNED_BYTE, buffer );

  va_end(a);
}

/*************** wglWindowClass interface **************/

struct wglWindowClass {
  wglWindowClass();

  void setContext( wglContext& cx );
  void clearContext( HWND hWnd );

  const char* name() { return wc.lpszClassName; }

  static wglWindowClass& instance() {
    static wglWindowClass* inst = new wglWindowClass;
    return *inst;
  }

  private:

  WNDCLASS wc;
  static std::map< HWND, wglContext* > contexts;

  static LRESULT WINAPI WindowProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );
};

/************* wglWindowClass implementation **************/

std::map< HWND, wglContext* > wglWindowClass::contexts;

void wglWindowClass::setContext( wglContext& cx ) {
  contexts[cx.hWnd] = &cx;
}

void wglWindowClass::clearContext( HWND hWnd ) {
  contexts[hWnd] = 0;
}

LRESULT WINAPI wglWindowClass::WindowProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
  wglContext* cx = contexts[hWnd];
  if (cx) 
    return cx->winMessage( hWnd, msg, wParam, lParam );
  
  return DefWindowProc(hWnd, msg, wParam, lParam);
}

wglWindowClass::wglWindowClass() {
  memset(&wc,0,sizeof(wc));
 
  wc.lpszClassName = "wglWindowC";
  wc.lpfnWndProc = WindowProc;
  wc.style = CS_OWNDC | CS_VREDRAW | CS_HREDRAW;
  wc.hInstance = GetModuleHandle(0);
  wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
  wc.hCursor = LoadCursor( NULL, IDC_ARROW );
  wc.hbrBackground = 0;
  wc.lpszMenuName = 0;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;

  RegisterClass(&wc);
}

/*************** wglContext implementation *************/

wglContext::wglContext() 
  : cls( wglWindowClass::instance() ),
    hWnd( 0 ),
    hdc( 0 ),
    hglrc( 0 ),
    lasthdc( 0 ),
    lastglrc( 0 ),

    mouseLocked( 0 ),
    buttonState( 0 ),
    buttonChanged( 0 ),
    mouseDelta(0,0,0),
    mousePos(0,0,0)
{
}

wglContext::~wglContext() {
  cleanup();
}

void wglContext::error(string where) {
  unsigned errorcode = GetLastError();


  char s[1024];
  FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | 80,  /* flags | max_width */
                 0,           /* source */
                 errorcode,   /* messageId */
                 0,           /* language */
                 s,           /* buffer */
                 sizeof(s)-1, /* buffer_size */
                 0            /* arguments */
               );

  error_message = where + ": " + s;
}

bool wglContext::changeWindow(const char* title, int x, int y, int width, int height) {
  if (!hWnd) return false;

  Vector org = origin();

  if ((x>=0 || y>=0 || width>=0 || height>=0) &&
      !MoveWindow( hWnd, x<0 ? org.x : x, y<0 ? org.y : y, width, height, TRUE))
    return false;

  if (title && !SetWindowText(hWnd, title))
    return false;

  return true;
}

bool wglContext::initWindow(const char* title, int x, int y, int width, int height) {
  cleanup();

  RECT screen;
  if (!SystemParametersInfo(SPI_GETWORKAREA, 0, &screen, 0))
    return false;

  if (x<0 && y<0 || x>screen.right || y>screen.bottom) {
    x = CW_USEDEFAULT;
    y = CW_USEDEFAULT;
  } else if (x<screen.left) {
    x = screen.left;
  } else if (y<screen.top) {
    y = screen.top;
  }

  if (x + width > screen.right) width = screen.right - x;
  if (y + height > screen.bottom) height = screen.bottom - y;

  hWnd = CreateWindow( cls.name(), 
                       title, 
                       WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                       x,y,width,height,
                       NULL, NULL, GetModuleHandle(0), NULL );

  if (!hWnd) { error("CreateWindow"); cleanup(); return false; }

  cls.setContext( *this );

  ShowWindow(hWnd, SW_SHOW);

  hdc = GetDC(hWnd);

  if (!hdc) { error("GetDC");cleanup(); return false; }

  PIXELFORMATDESCRIPTOR pfd = { 
      sizeof(PIXELFORMATDESCRIPTOR),   // size of this pfd 
      1,                               // version number 
      PFD_DRAW_TO_WINDOW |             // output to screen (not an image) 
      PFD_SUPPORT_OPENGL |             // support OpenGL 
      PFD_DOUBLEBUFFER,                // double buffered 
      PFD_TYPE_RGBA,                   // RGBA type 
      24,                              // 24-bit color depth 
      0, 0, 0, 0, 0, 0,                // color bits ignored 
      0,                               // no alpha buffer 
      0,                               // shift bit ignored 
      0,                               // no accumulation buffer 
      0, 0, 0, 0,                      // accum bits ignored 
      32,                              // 32-bit z-buffer 
      0,                               // no stencil buffer 
      0,                               // no auxiliary buffer 
      PFD_MAIN_PLANE,                  // main layer 
      0,                               // reserved 
      0, 0, 0                          // layer masks ignored 
  };

  int pixelFormat = ChoosePixelFormat( hdc, &pfd );

  if (!pixelFormat) { error("ChoosePixelFormat"); cleanup(); return false; }

  if (!DescribePixelFormat( hdc, pixelFormat, sizeof(pfd), &pfd )) {
    error("DescribePixelFormat"); cleanup(); return false; 
  }

  if (!SetPixelFormat( hdc, pixelFormat, &pfd )) {
    error("SetPixelFormat"); cleanup(); return false; 
  }

  hglrc = wglCreateContext(hdc);

  if (!hglrc) { error("wglCreateContext"); cleanup(); return false; }

  // Init fonts:

  makeCurrent();
  SelectObject (hdc, GetStockObject (SYSTEM_FONT));
  sysFont = glGenLists(256);
  wglUseFontBitmaps(hdc, 0, 256, sysFont);
  makeNotCurrent();

  // Init mouse cursor:

  POINT p; 
  GetCursorPos(&p);
  mousePos.x = p.x;
  mousePos.y = p.y;

  return true;
}

bool wglContext::isOpen() {
  return hWnd != 0;
}

void wglContext::cleanup() {
  if (hglrc) {
    wglDeleteContext(hglrc);
    hglrc = 0;
  }
  if (hdc) {
    ReleaseDC(hWnd, hdc);
    hdc = 0;
  }
  if (hWnd) {
    DestroyWindow(hWnd);
    hWnd = 0;
  }
  mustPaint = false; 
}

void wglContext::lockMouse() {
  if (mouseLocked || !hWnd) return;

  POINT p; p.x = 0; p.y = 0;
  ScreenToClient(hWnd,&p);
  RECT r;
  GetClientRect(hWnd,&r);
  r.left -= p.x; r.right -= p.x;
  r.top -= p.y; r.bottom -= p.y;
  ClipCursor(&r);

  ShowCursor(false);
  mouseLocked = true;

  GetCursorPos(&p);
  mousePos.x = p.x;
  mousePos.y = p.y;
}

void wglContext::unlockMouse() {
  if (!mouseLocked) return;

  ShowCursor(true);
  ClipCursor(0);
  mouseLocked = false;
}

Vector wglContext::getMousePos() {
  if (!hWnd) return Vector(0,0,0);

  RECT r;
  GetClientRect(hWnd, &r);
  POINT p; p.x = mousePos.x; p.y = mousePos.y;
  ScreenToClient(hWnd, &p);

  return Vector( 
    clamp( float(p.x) / r.right, 0, 1 ),
    clamp( float(p.y) / r.bottom, 0, 1 ),
    mousePos.z );
}

Vector wglContext::getMouseDelta() {
  if (!hWnd) return Vector(0,0,0);

  RECT r;
  GetClientRect(hWnd, &r);

  Vector d( mouseDelta.x / r.right,
         mouseDelta.y / r.bottom,
         mouseDelta.z );

  mouseDelta = Vector(0,0,0);

  return d;
}

int wglContext::getMouseButtons() {
  return buttonState;
}

int wglContext::getMouseButtonsChanged() {
  unsigned ch = buttonChanged;
  buttonChanged = 0;
  return ch;
}

LRESULT wglContext::winMessage( HWND, UINT msg, WPARAM wParam, LPARAM lParam) {
  switch( msg ) {
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_MBUTTONDOWN:
    case WM_MBUTTONUP:
    case WM_MOUSEMOVE: {
      int newState =  ( (wParam&MK_LBUTTON)!=0) +
                      (((wParam&MK_RBUTTON)!=0)<<1) +
                      (((wParam&MK_MBUTTON)!=0)<<2);    

      POINT cursor;
      GetCursorPos(&cursor);

      Vector delta( cursor.x - mousePos.x,
                    cursor.y - mousePos.y,
                    0 );

      buttonChanged |= (buttonState ^ newState);
      buttonState = newState;
      mouseDelta = mouseDelta + delta;

      if (mouseLocked) {
        if (mousePos.x != cursor.x ||
            mousePos.y != cursor.y)
          SetCursorPos( mousePos.x, 
                        mousePos.y );
      } else {
        mousePos.x = cursor.x;
        mousePos.y = cursor.y;
      }

      break;
      }

    case WM_PAINT:
      // force display refresh
      mustPaint = true;
      makeCurrent();
      glClear(GL_COLOR_BUFFER_BIT);
      swapBuffers();

      break;

    case WM_DESTROY:
      cls.clearContext( hWnd );
      hWnd = 0;
      return 0;
  }
  
  return DefWindowProc( hWnd, msg, wParam, lParam );
}

void wglContext::makeCurrent() {
  lasthdc = wglGetCurrentDC();
  lastglrc = wglGetCurrentContext();
  wglMakeCurrent( hdc, hglrc );
}

void wglContext::makeNotCurrent() {
  wglMakeCurrent( lasthdc, lastglrc );
  lasthdc = 0; lastglrc = 0;
}

void wglContext::swapBuffers() {
  SwapBuffers(hdc);
}

Vector wglContext::origin() {
  RECT r;
  GetWindowRect(hWnd, &r);
  return Vector(r.left,r.top);
}

Vector wglContext::corner() {
  RECT r;
  GetWindowRect(hWnd, &r);
  return Vector(r.right,r.bottom); 
}

int wglContext::width() {
  RECT r;

  if (!hWnd) return 0;
  GetClientRect(hWnd, &r);
  return r.right;
}

int wglContext::height() {
  RECT r;

  if (!hWnd) return 0;
  GetClientRect(hWnd, &r);
  return r.bottom;
}
