#include "xgl.h"

#define clamp(x, a, b) (__max(a, __min(x,b) ))

int xglContext::sysFont = -1;

void xglContext::glPrintf(int font, char* format, ...) {
  if (font<0) return;
  va_list a;
  va_start(a,format);
  
  char buffer[1024];
  vsprintf(buffer, format, a);
  int len = strlen(buffer);
  glListBase(font);
  glCallLists( len, GL_UNSIGNED_BYTE, buffer );

  va_end(a);
}

/*************** xglContext implementation *************/

xglContext::xglContext() 
 : window(NULL),
   area(NULL),
   buttonState(0),
   buttonsChanged(0),
   mouseLocked(false),
   mustPaint(false)
{
}

xglContext::~xglContext() {
  cleanup();
}

gint xglContext::configure_cb(GtkWidget *widget,
                              GdkEventConfigure *event,
                              gpointer data) {
  xglContext *c = (xglContext *)data;
  c->wwidth = event->width;
  c->wheight = event->height;

  return TRUE;
}

gint xglContext::motion_notify_cb(GtkWidget *widget,
                                         GdkEventMotion *event,
                                         gpointer data) {
  xglContext *c = (xglContext *)data;
  c->mousePos.x = event->x;
  c->mousePos.y = event->y;

  return TRUE;
}

gint xglContext::delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data) {
  xglContext *c = (xglContext *)data;
  c->cleanup();
  return TRUE;
}

gint xglContext::button_press_cb(GtkWidget *widget, GdkEventButton *event,
                                 gpointer data) {
  xglContext *c = (xglContext *)data;

  /* Dave and X don't agree about button labels */
  if (event->button == 2)
    event->button = 3;
  else if (event->button == 3)
    event->button = 2;

  c->buttonState |= (0x1 << (event->button - 1));
  c->buttonsChanged |= (0x1 << (event->button -1));

  return TRUE;
}

gint xglContext::button_release_cb(GtkWidget *widget, GdkEventButton *event,
                                 gpointer data) {
  xglContext *c = (xglContext *)data;

  /* Dave and X don't agree about button labels */
  if (event->button == 2)
    event->button = 3;
  else if (event->button == 3)
    event->button = 2;

  c->buttonState &= ~(0x1 << (event->button - 1));
  c->buttonsChanged |= (0x1 << (event->button -1));

  return TRUE;
}

bool xglContext::changeWindow(const char* title, int x, int y, int width, int height) {
  return false;
}

bool xglContext::initWindow(const char* title, int x, int y, int width, int height) {
  cleanup();

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), title);
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);

  int attrlist[] = {
    GDK_GL_RGBA,
    GDK_GL_DOUBLEBUFFER,
    GDK_GL_DEPTH_SIZE, 12,
    GDK_GL_NONE
  };

  if (!(area = GTK_WIDGET (gtk_gl_area_new(attrlist)))) {
    error_message = "Unable to create OpenGL display widget";
    cleanup();
    return false;
  }

  gtk_container_add (GTK_CONTAINER (window), area);

  gtk_widget_set_events(GTK_WIDGET (area),
                        GDK_EXPOSURE_MASK |
                        GDK_BUTTON_PRESS_MASK |
                        GDK_BUTTON_RELEASE_MASK |
                        GDK_POINTER_MOTION_MASK);

  gtk_signal_connect(GTK_OBJECT(window), "configure_event",
                     GTK_SIGNAL_FUNC(&xglContext::configure_cb),
                     (gpointer)this);
  gtk_signal_connect(GTK_OBJECT(area), "motion_notify_event",
                     GTK_SIGNAL_FUNC(&xglContext::motion_notify_cb),
                     (gpointer)this);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     GTK_SIGNAL_FUNC(&xglContext::delete_cb), (gpointer)this);
  gtk_signal_connect(GTK_OBJECT(area), "button_press_event",
                     GTK_SIGNAL_FUNC(&xglContext::button_press_cb),
                     (gpointer)this);
  gtk_signal_connect(GTK_OBJECT(area), "button_release_event",
                     GTK_SIGNAL_FUNC(&xglContext::button_release_cb),
                     (gpointer)this);

  gtk_gl_area_size(GTK_GL_AREA(area), width, height);
  wwidth = width;
  wheight = height;

  // Init fonts:

  /*
  makeCurrent();
  SelectObject (hdc, GetStockObject (SYSTEM_FONT));
  sysFont = glGenLists(256);
  wglUseFontBitmaps(hdc, 0, 256, sysFont);
  makeNotCurrent();
  */

  // Init mouse cursor:

  /*
  POINT p; 
  GetCursorPos(&p);
  mousePos.x = p.x;
  mousePos.y = p.y;
  */

  gtk_widget_show_all (window);

  return true;
}

bool xglContext::isOpen() {
  return window != NULL;
}

void xglContext::cleanup() {
  if (area) {
    gtk_widget_destroy (area);
    area = NULL;
  }
  if (window) {
    gtk_widget_destroy (window);
    window = NULL;
  }
  wwidth = 0;
  wheight = 0;
  mustPaint = false; 
}

void xglContext::lockMouse() {
}

void xglContext::unlockMouse() {
}

Vector xglContext::getMousePos() {
  if (!window) return Vector(0,0,0);

  Vector tmp = mousePos;
  tmp.x /= wwidth;
  tmp.y /= wheight;

  return tmp;
}

Vector xglContext::getMouseDelta() {
  Vector tmp = mousePos - oldMousePos;
  oldMousePos = mousePos;
  tmp.x /= wwidth;
  tmp.y /= wheight;

  return tmp;
}

int xglContext::getMouseButtons() {
  return buttonState;
}

int xglContext::getMouseButtonsChanged() {
  int c = buttonsChanged;
  buttonsChanged = 0;
  return c; 
}

void xglContext::makeCurrent() {
  gtk_gl_area_make_current (GTK_GL_AREA (area));
}

void xglContext::makeNotCurrent() {
}

void xglContext::swapBuffers() {
  gtk_gl_area_swapbuffers (GTK_GL_AREA (area));
}

Vector xglContext::origin() {
  /* FIXME */
  return Vector(0,0,0);
}

Vector xglContext::corner() {
  return origin() + Vector(wwidth,wheight);
}

int xglContext::width() {
  return wwidth;
}

int xglContext::height() {
  return wheight;
}
