cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c PROGRAM FINSIM.FOR
c
c This is the program simulating horizontal component of shear wave from 
c a finite fault. The large fault plane is subdivided into rectangular 
c elements, each element (subfault) is assigned an w-squared spectrum, and 
c delayed contributions from all subfaults are summed up in the time domain. 
c In generating subfault accelerograms, a stochastic point-source technique 
c ("band-limited white noise method") is followed (Boore, 1983; Boore and 
c Atkinson, 1987). The units of simulated accelerogram are cm/sec**2. The 
c accelerogram, simulation parameters, and pseudo-acceleration response 
c spectrum are written into specified ascii files
c
c The program is limited to simulations of frequencies of 0.25 Hz and
c higher. The subfault size should not be less than appr. 5 km to make 
c simulations reasonably subfault-size independent
c
c The user should provide fault discretization and slip distribution models. 
c The input data should be prepared in a file. The program will ask for the
c name of input file. An example input file is enclosed
c
c All subroutines are attached
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c      Authors: Igor A. Beresnev and Gail M. Atkinson (September'97)    c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c The dimension of the array f1, containing simulated finite-fault 
c accelerogram, is 10000. Data are sampled at dt sec. The leng1-point Fast 
c Fourier Transform is used. dt and leng1 are user-specified. Maximum nl 
c (number of subfaults along fault strike, Figure 1) is 50. Maximum nw 
c (number of subfaults along dip) is 50
c
	dimension f(4096),f1(10000),f2(10000)
	dimension slip(50,50),slip2(50,50),freq(100),psa(100),psa1(100)
	character finp*30,facc*30,fpar*30,fpsa*30,fresp1*30,fresp2*30
	character comm*70
	common /par/ leng,leng1,dt
        common /sub/ ame,f0,rho,beta,ikap,fmax,Q0,eta,rise,rmin,rd1,
     *   rd2,durmin,b1,b2,b3,igeom,rg1,rg2,pow1,pow2,pow3,iwind,
     *   namp1,namp2
	common /seed/ islip
	common /fnames/ fresp1,fresp2
	data psa1 /100*0./
c
c leng is maximum number of data points in simulated accelerogram
	leng=10000
	pi=3.141593
c y is the fraction of rupture propagation velocity to shear-wave velocity
	y=0.80
c
c reading input parameters (see Figure 1 for geometry)
c A rectangular fault with horizontal upper and lower sides is assumed. 
c Origin O is always placed at the upper corner of the fault, opposite to 
c strike direction. A convention defining strike direction is that fault 
c always dips to the right of it (Aki and Richards, "Quantitative Seismology", 
c Fig.4.13). alat1 and alon1 are latitude and longitude of surface projection
c (O1) of the origin (see Aerial View in Figure 1)
c
c comm        - comment to be written as first line in output parameter file
c               (70 characters)
c fi1         - fault strike (degrees) (0 to 360)
c dip1        - fault dip measured from horizontal (degrees) (0 to 90)
c h	      - depth to upper edge of the fault (km)
c alat1,alon1 - latitude and longitude of the origin (degrees). Convention
c               for sign of coordinates: '+' for North and East, '-' for South 
c               and West
c alat2,alon2 - latitude and longitude of the observation point (degrees)
c rleng       - fault length along strike (km)
c wid         - fault width downdip (km)
c nl          - number of subfaults along strike
c nw          - number of subfaults along dip
c i0,j0       - the indices of the subfault that contains the hypocenter
c	        i and j increase along strike and downdip, respectively, 
c	        starting from 1 at origin (Figure 1)
c leng1       - number of points in Fast Fourier Transform used to calculate
c               subsource accelerograms. Must be a power of two (maximum of
c               4096 is allowed). If not sufficient, the program will give
c               you a warning
c dt          - time-series sampling interval (sec)
c beta        - crustal shear-wave velocity (km/sec) (e.g., average value
c               for depth of event)
c rho         - crustal density (g/cm**3)
c amag        - moment magnitude of modeled (target) event 
c sfact       - factor controlling strength of subfault radiation. Enter
c               1. if you wish to model a "standard" event. Increase or 
c               decrease this value to model "unusual" fast or slow events, 
c               respectively. The amplitudes will change proportionally to
c               sfact**2. Physically, this will correspond to a change in 
c               maximum slip velocity on the fault proportionally to first 
c               power of sfact. Recommended values of sfact are between 0.5
c               and 2.0
c stress      - "stress" parameter (bars). Controls subfault moment and total 
c               number of subsources summed. To avoid insufficient number of
c               "active" soubsources, we recommend it be always kept at 
c               "average" value of 50 bars. Lower values are possible; higher 
c               values should not generally be used
c ikap        - 0 or 1 for "fmax" and "kappa" low-pass filters, respectively
c fmax        - cutoff frequency of "fmax" filter (if ikap=0) or parameter 
c               "kappa" (if ikap=1)
c Q0,eta      - parameters of attenuation model Q(f)=Q0*f**eta
c igeom,rg1,  
c rg2,pow1,
c pow2,pow3   - geometric attenuation model. If igeom=0, 1/r model is
c               applied. If igeom=1, "trilinear" form is used: r**pow1 
c               (r <= rg1),r**pow2 (rg1 < r <= rg2), r**pow3 (r > rg2). rg1
c               and rg2 are in km
c rmin,rd1,
c rd2,durmin,
c b1,b2,b3    - parameters of "trilinear" distance-dependent duration model 
c               used for subevents. A constant term durmin is added to source 
c               rise time at r<=rmin. Duration increases linearly with distance 
c               as dur(r) = b*r with slopes b1, b2, b3 at distances between 
c               rmin and rd1, rd1 and rd2, and beyond rd2, respectively. r's 
c               are in km, durmin is in sec. 
c               A model rmin = 50 km, durmin = 1.4 sec, b1 = 0.07, and rd1, rd2 
c               set to "infinite" values is approximately valid for subduction-
c               zone events ("one-slope" model) (Atkinson, BSSA, no.5/95; 
c               Beresnev and Atkinson, BSSA, no.1/97).
c               For eastern North America, one might use rmin = 10 km, 
c               rd1 = 70 km, rd2 = 130 km, durmin = 0 sec, b1 = 0.16, 
c               b2 = -0.03, b3 = 0.04 (Atkinson and Boore, BSSA, no.1/95).
c               To keep duration equal to source rise time at all distances,
c               simply set durmin = 0 sec and rmin at "infinite" value
c iwind       - if 0, a cosine-tapered boxcar window will be applied to 
c               subfault time series; if 1, a Saragoni-Hart window will be
c               applied. Using 1 produces more realistic-looking accelerograms 
c               when total number of subfaults summed is relatively low, 
c               e.g., for target magnitudes of about 6-7
c nfreq,
c freq1,freq2 - response spectrum is calculated at nfreq logarithmically-
c               spaced frequencies between freq1 and freq2 (Hz). freq1 
c               should be no lower than 0.25 Hz. nfreq cannot exceed 100
c facc,fpar,
c fpsa        - names of ascii files in which to save acceleration time 
c               history, simulation parameters, and average response 
c               spectrum, respectively
c namp1,fresp1- namp1 determines if spectra should be amplified by site 
c               amplification function. For convenience, user may specify two 
c               separate amplifications: e.g., one due to generic crustal model 
c               and one due to local site response. Application of second 
c               function is controlled by namp2,fresp2 below. If namp1 is set
c               to 0, no amplification is introduced. If you enter a non-zero
c               value, prepare an ascii file and provide its name as fresp1.
c               The file should have two columns, frequency (real) and
c               amplification (real), separated by space or tab. The file
c               should contain namp1 lines; maximum namp1 is 100. Program
c               will interpolate. For frequencies to the left and right of
c               lowest and highest values provided, amplification will be
c               assumed constant. For example, if you provide response
c               function between 1 and 10 Hz, amplification below 1 Hz will 
c               be assumed as at 1 Hz, and above 10 Hz as at 10 Hz. Put any
c               dummy file name if namp1 = 0 
c namp2,fresp2- same as namp1,fresp1 for second amplification function
c ntrial,damp - number of simulations to calculate average response 
c               spectrum (ntrial). A time series saved is from first
c               simulation. damp is percentage of critical damping for
c               calculation of response spectrum
c islip       - enter 0 if you use your own slip distribution model. Then
c               provide slip(i,j) below. Enter any other integer number
c               if you don't want to enter your model. The program 
c               will skip reading array slip(i,j) and will generate random 
c               normally distributed slip with standard deviation equal to 
c               mean. islip will be the seed of random-number generator
c slip(i,j)   - array of slip weights on individual subfaults in any relative 
c               units. Allows to specify inhomogeneous slip. Slip weights need 
c               not sum to unity; program will do normalization. Subfault 
c               moments will be weighted accordingly so that total moment is
c               conserved. First enter upper horizontal row, then second
c               horizontal row, and so on to the bottom (Figure 1). The array
c               is (nl x nw). If slip is homogeneous, enter same number for
c               all subfaults. Because subfault moments are discrete, the 
c               given slip distribution slip(i,j) will be mimicked but not
c               exactly reproduced. The program will return the actual slips 
c               on subfaults in meters in an output parameter file
c
c dl and dw below are subfault length and width, respectively. nl and nw
c should be specified in such a way that (dl+dw)/2 is no less than 5 km
c
	write (6,*) 'input file name?'
	read (5,'(a)') finp
	open (2,form='formatted',file=finp)
	read (2,'(a)') comm
	read (2,*) fi1,dip1,h
	read (2,*) alat1,alon1,alat2,alon2
	read (2,*) rleng,wid,nl,nw
	dl=rleng/nl
	dw=wid/nw
	read (2,*) i0,j0
	read (2,*) leng1,dt
	read (2,*) beta,rho
	read (2,*) amag,sfact,stress,ikap,fmax
	read (2,*) Q0,eta
	read (2,*) igeom,rg1,rg2,pow1,pow2,pow3
	read (2,*) rmin,rd1,rd2,durmin,b1,b2,b3
	read (2,*) iwind
	read (2,*) nfreq,freq1,freq2
	read (2,'(a)') facc
	read (2,'(a)') fpar
	read (2,'(a)') fpsa
	read (2,*) namp1
	read (2,'(a)') fresp1
	read (2,*) namp2
	read (2,'(a)') fresp2
	read (2,*) ntrial,damp
	read (2,*) islip
	if (islip.eq.0) read (2,*) ((slip(i,j),i=1,nl),j=1,nw)
	close (2)
c check allowable values for leng1
	icheck=0
	do 18 i=1,12
	ncheck=2**i
	if (leng1.eq.ncheck) icheck=1
18	continue
	if (icheck.ne.1) then
	write (6,*)  'ERROR. leng1 must be a power of 2. Max is 4096'
	stop
	endif
c v is rupture propagation velocity
	v=y*beta
c z is fundamental constant relating corner frequency of w-squared spectrum
c to subfault size. It is linked to maximum slip velocity on fault (vmax)
c as vmax = (2*y*z/e)*stress/(rho*beta), where e is base of natural 
c logarithm. The amplitude of subfault radiation is proportional to z**2.
c The "standard" value of z is 1.68 (Beresnev and Atkinson, BSSA, no.1/97,
c p.70). Input parameter sfact controls radiation strength. sfact = 1. 
c corresponds to "standard" condition
	z=sfact*1.68
c convert magnitude to target moment (aml)
	aml=10.**(1.5*amag+16.05)
c
c generate random slip distribution if islip is not zero
	if (islip.ne.0) then
	do 13 i=1,nl
	do 14 j=1,nw
	slip(i,j)=ggnqf()+1.
	if (slip(i,j).lt.0.) slip(i,j)=0.
14	continue
13	continue
	endif
c
c calculate moments of subfaults. First calculate the total of all weights
c (slip1), then moment per unit weight (elem), then individual moments and 
c put them back into array slip(i,j) 
	slip1=0.
	do 6 i=1,nl
        do 5 j=1,nw
	slip1=slip1+slip(i,j)
5	continue
6	continue
	elem=aml/slip1
	do 15 i=1,nl
	do 16 j=1,nw
	slip(i,j)=slip(i,j)*elem
16	continue
15	continue
c
c obtain values of fi2 and R from latitudes and longitudes
	call geom (alat1,alon1,alat2,alon2,R,fi2)
c convert degrees to radians
	fi1=pi/180.*fi1
	fi2=pi/180.*fi2
	dip=90.-dip1
	dip=pi/180.*dip
	al=(dl+dw)/2. 
c warning if subfault size too small
	if (al.lt.5.) then 
	write (6,*)
        write (6,*) 'WARNING:'
	write (6,*) 'Average subfault size is less than 5 km.'
	write (6,*) 'Simulation will be subfault-size dependent.'
	write (6,*) 'Consider reducing number of subfaults along'
	write (6,*) 'strike and dip (nl and nw)'
	pause 'Press <Enter> to continue this simulation'
	endif
c
c estimate average number of triggerings on individual subsource
	na=nint(aml/(rleng*wid*stress*al*1.e21))
c warn if zero 
	if (na.eq.0) then
	write (6,*)
	write (6,*) 'WARNING:' 
	write (6,*) 'Total number of subsources summed may be too'
	write (6,*) 'small. If you see "gaps" on the accelerogram,'
	write (6,*) 'increase the number of subfaults along strike'
	write (6,*) 'and dip (nl and nw), or increase the target'
	write (6,*) 'magnitude (without changing fault dimensions)'
	pause 'Press <Enter> to continue this simulation'
	endif
c calculate subfault rise time (rise), subfault corner frequency (f0),
c subfault moment (ame) in dyne*cm (Beresnev and Atkinson, BSSA, no.1/97, 
c eqs. 13,14,16)
c usub is subfault displacement in m given by its moment
	rise=al/v
	f0=(y*z/pi)*(beta/al)
	ame=1.e21*stress*al**3.
	usub=1.e-22*ame/(rho*beta**2.*dl*dw)
	ntrig=0
	dmin=1.e4
	dmax1=0.
	dmax2=0.
c
c calculate miminum time delay (dmin) for subfault radiation to reach 
c observation point. Calculate estimated total duration (dmax) of finite-
c fault accelerogram. Numbers i,j define current subfault (Figure 1)
		do 8 i=1,nl
		   do 9 j=1,nw
c get distance r1 from center of subfault to obs. point (eq.1, Figure 1)
		   call dis (R,r1,h,fi1,fi2,dip,dl,dw,i,j)
c get time delay due to rupture propagation to the center of subfault 
c (eq.2, Figure 2)
		   del=sqrt((dl*(i-i0))**2.+(dw*(j-j0))**2.)/v
c get total time delay (del)
		   del=del+r1/beta
		   if (del.lt.dmin) dmin=del
		   if (del.gt.dmax1) dmax1=del
		   call durcmp (r1,dummy)
		   if (dummy.gt.dmax2) dmax2=dummy
9		   continue
8		continue
c
c dmax1 is maximum delay due to geometry. dmax2 is maximum soubsource
c duration. One second is added to dmax for sufficiency. Static time shift
c (dmin) is subtracted, but constant shift of 2 sec is retained (see do...2 
c loop below). lengts is number of data points in simulated finite-fault 
c accelerogram to be saved in output file (no more than leng)
	dmax=dmax1+dmax2-(dmin-2.)+1.
	lengts=dmax/dt
	if (lengts.gt.leng) lengts=leng
c
c ***** now do actual summation
c outer loop (do...11) is active when average response spectrum is requested
c
	do 11 iresp=1,ntrial
	write (6,10) iresp
10	format (/' *** trial',i4)
		do 12 i=1,leng
		f1(i)=0.
12		continue
c
c ***************** main summation is done below ***********************
c sum up radiation from all subfaults (loop 1)
c
		do 1 i=1,nl
		   do 2 j=1,nw
c n is the number of subsource triggerings needed to match its moment
		   n=nint(slip(i,j)/ame)
c put actual slips in m in slip2 (i,j) 
		   slip2(i,j)=n*usub
		   if (n.eq.0) go to 2
		   call dis (R,r1,h,fi1,fi2,dip,dl,dw,i,j)
		   del=sqrt((dl*(i-i0))**2.+(dw*(j-j0))**2.)/v
c subtract minimum delay to eliminate static time shift, but retain constant
c static shift of 2 s
		   del=(del+r1/beta)-dmin+2.
c compute stochastic accelerogram from a subfault
		   call sei (f,r1)
			    do 4 isrc=1,n
c count cumulative number of subsources summed
			    ntrig=ntrig+1
c subsource accelerogram is randomly delayed to simulate complexity
c in slip process
		            del1=del+(isrc-1+urand(islip))*rise
c shift generated time series by del1 and return in array f2
		            call shift (f,f2,del1)
c add current accelerogram to total wave field
					do 3 k=1,leng
					f1(k)=f1(k)+f2(k)
3					continue
4			    continue
2		   continue
1		continue
************************** end of summation ****************************
c
	if (ntrig.eq.0) then
	write (6,*)
	write (6,*) 'PROGRAM STOPS: total number of subsources summed'
	write (6,*) 'is zero. Try new simulation. Increase number of'
        write (6,*) 'subfaults along strike and dip (nl and nw), or'
	write (6,*) 'slightly increase target magnitude without change'
	write (6,*) 'in fault dimensions'
	stop
	endif
c accelerogram and model parameters are saved only once
	if (iresp.eq.1) then
c aml and amag now are actual total moment and magnitude achieved
	aml=ame*ntrig
	amag=alog10(aml)*2./3.-10.7
c find absolute maximum acceleration in simulated time history
	amax=0.
		do 7 i=1,leng	
		if (abs(f1(i)).gt.amax) amax=abs(f1(i))
7		continue
c save accelerogram and model parameters
	call writa (facc,fpar,comm,f1,amax,lengts)
	call writp (fpar,comm,180./pi*fi1,dip1,h,alat1,alon1,alat2,
     *  alon2,rleng,wid,nl,nw,dl,dw,i0,j0,amag,sfact,stress,ntrig,slip2)
	endif
c
c calculate individual response spectrum at nfreq frequencies between freq1 
c and freq2
	dur=(leng-1)*dt-5. 
	call rscomp (f1,dt,dur,damp,freq1,freq2,nfreq,freq,psa)
c calculate average response spectrum
	do 17 i=1,nfreq
	psa1(i)=psa1(i)+psa(i)/ntrial
17	continue
c
11	continue
c save average response spectrum
	call writr (fpsa,fpar,comm,freq,psa1,nfreq,ntrial,damp)
	stop
	end
c
c
	subroutine shift (f,f2,del)
	dimension f(*),f2(*)
	common /par/ leng,leng1,dt
		do 5 i=1,leng
		f2(i)=0.
5		continue
	ish=int(del/dt)+1
	if (ish.gt.leng+1) return
	iend=ish-1+leng1
	if (iend.gt.leng) iend=leng
	j=0
		do 2 i=ish,iend
		j=j+1
		f2(i)=f(j)
2		continue
	return
	end
c
c
	subroutine dis (R,r1,h,fi1,fi2,dip,dl,dw,i,j)
c	computes distance r1 from center the of subfault (i,j)
c	to observation point using formula (1) (Figure 1)
	t1=R*cos(fi2-fi1)-(2.*i-1)*dl/2.
	t2=R*sin(fi2-fi1)-(2.*j-1)*dw/2.*sin(dip)
	t3=-h-(2.*j-1)*dw/2.*cos(dip)
	r1=sqrt(t1**2.+t2**2.+t3**2.)
	return
	end
c
c
	real function urand (iy) 
c	uniform random number generator. Values of urand are returned
c	in the interval (0,1) (G.E. Forsythe, M.A. Malcolm, C.B. Moler,
c	"Computer Methods for Mathematical Computations", Prentice-Hall, Inc.)
	integer iy,ia,ic,itwo,m2,m,mic
	double precision halfm,datan,dsqrt
	real s
	data m2/0/,itwo/2/
	if(m2.ne.0) go to 20
	m=1
10      m2=m
	m=itwo*m2
	if (m.gt.m2) go to 10
	halfm=m2
	ia=8*idint(halfm*datan(1.D0)/8.D0)+5
	ic=2*idint(halfm*(0.5D0-dsqrt(3.D0)/6.D0))+1
	mic=(m2-ic)+m2
	s=0.5/halfm
20      iy=iy*ia
        if(iy.gt.mic)iy=(iy-m2)-m2
	iy=iy+ic
        if(iy/2.gt.m2)iy=(iy-m2)-m2
	if(iy.lt.0)iy=(iy+m2)+m2
	urand=float(iy)*s
	return
	end
c
c
      subroutine sei (seism,r1)
c     calculates accelerogram from individual subfault by using band-limited
c     white noise method (Boore, BSSA, no.6/83). An w-squared spectrum is
c     assumed
      dimension seism(*)
      complex s(4096)
      common /par/ leng,leng1,dt
      common /sub/ ame,f0,rho,beta,ikap,fmax,Q0,eta,rise,rmin,rd1,
     *   rd2,durmin,b1,b2,b3,igeom,rg1,rg2,pow1,pow2,pow3,iwind,
     *   namp1,namp2
      data iflag /0/
c
      if (iflag.eq.1) go to 1
c specify point-source constants
      iflag=1
      pi=3.141593
      tmax=leng1*dt
      df=1./tmax
      taper=0.02
      prtitn=1./sqrt(2.)
      rtp=0.55
      fs=2.
c define spectral constant, for r1=1 km
      const=prtitn*rtp*fs*(1.e-20)/(4.*pi*rho*beta**3.)
      nnyq=leng1/2+1
c
c calculate duration of subsource time history at distance r1
1     call durcmp (r1,dur)
      if (dur.lt.0.) stop 'Duration is negative. Check duration model'
      ndur=dur/dt
      ntaper=int(taper*ndur)
      nstop=ndur+ntaper+ntaper
      if(nstop.gt.leng1) then
      write (6,*) 'PROGRAM STOPS: subsource duration too long.'
      write (6,*) 'Options to fix the problem:'
      write (6,*) '   1) Increase leng1 or dt, or'
      write (6,*) '   2) Reduce distance to observation point, or'
      write (6,*) '   3) Change parameters of duration model'
      stop
      endif
c
      do 110 i=1,leng1
      seism(i)=0.
      s(i)=cmplx(0.,0.)
110   continue
c
c generate Gaussian white noise with zero mean, unit variance. Noise is 
c tapered.
c
      do 120 i=1,nstop
      if (iwind.eq.0) call window(i,1,nstop,ntaper,wind)
      if (iwind.eq.1) call wind2 (i,1,nstop,ntaper,dur,0.2,0.2,wind)
      s(i)=wind*cmplx(ggnqf(),0.)
120   continue
c
c transform to frequency domain
c
      call fork (leng1,s,-1.)
c
c multiply spectrum by an w-squared spectral shape after normalizing to 
c square of unit spectral amplitude
c
      call avg (nnyq,s,avamp2)
      avamp=sqrt(avamp2)
        do 130 i=1,nnyq
        f=(i-1)*df
        s(i)=shape(f,r1,const)*s(i)/avamp
        if (i.ne.1) s(leng1+2-i)=conjg(s(i))
130     continue
	s(nnyq)=cmplx(real(s(nnyq)),0.)
c
c     transform back to time domain
      call fork(leng1,s,1.)
      afact=sqrt(float(leng1))/tmax
        do 140 i=1,leng1
        seism(i)=afact*real(s(i))
140     continue
      return
      end
c
c
      subroutine durcmp (r1,dur)
c     computes duration dur of soubsource time histories
      common /sub/ ame,f0,rho,beta,ikap,fmax,Q0,eta,rise,rmin,rd1,
     *   rd2,durmin,b1,b2,b3,igeom,rg1,rg2,pow1,pow2,pow3,iwind,
     *   namp1,namp2
      if (r1.le.rmin) dur=rise+durmin
      if (r1.gt.rmin.and.r1.le.rd1) dur=rise+durmin+b1*(r1-rmin)
      if (r1.gt.rd1.and.r1.le.rd2) 
     *   dur=rise+durmin+b1*(rd1-rmin)+b2*(r1-rd1)
      if (r1.gt.rd2) 
     *   dur=rise+durmin+b1*(rd1-rmin)+b2*(rd2-rd1)+b3*(r1-rd2)
	return
	end
c
c
      function ggnqf
c     generates Gaussian white noise by approx. method of Ross (1985)
      common /seed/ islip
      usum=0.
      do 20 j=1,12
      usum=usum+urand(islip)
20    continue
      ggnqf=usum-6.
      return
      end
c
c
      subroutine avg(n,s,avamp2)
c     calculates average squared amplitude spectrum
      complex s
      dimension s(*)
      sum=0.
      do 10 j=1,n
      amp=cabs(s(j))
      sum=sum+amp*amp
10    continue
      avamp2=sum/float(n)
      return
      end
c
c
      subroutine fork(lx,cx,signi)
c     Fast Fourier Transform
      complex cx,carg,cexp,cw,ctemp
      dimension cx(lx)
      j=1
      sc=sqrt(1./lx)
      do 5 i=1,lx
          if(i.gt.j) go to 2
          ctemp=cx(j)*sc
          cx(j)=cx(i)*sc
          cx(i)=ctemp
2         m=lx/2
3         if(j.le.m) go to 5
            j=j-m
            m=m/2
            if(m.ge.1) go to 3
5     j=j+m
      l=1
6     istep=2*l
      temp=3.141593*signi/l
      do 8 m=1,l
          carg=(0.,1.)*temp*(m-1)
          cw=cexp(carg)
          do 8 i=m,lx,istep
               ctemp=cw*cx(i+l)
               cx(i+l)=cx(i)-ctemp
8     cx(i)=cx(i)+ctemp
      l=istep
      if(l.lt.lx) go to 6
9     return
      end
c
c
      real function shape (f,r1,const)
c     Computes amplitude value of w-squared spectrum at distance r1.
c     Source terms correspond to r1=1 km
      dimension afreq1(100),afreq2(100),amp1(100),amp2(100)
      character fresp1*30,fresp2*30
      common /sub/ ame,f0,rho,beta,ikap,fmax,Q0,eta,rise,rmin,rd1,
     *   rd2,durmin,b1,b2,b3,igeom,rg1,rg2,pow1,pow2,pow3,iwind,
     *   namp1,namp2
      common /fnames/ fresp1,fresp2
      data iflag/0/
c
c read site-response functions
      if (iflag.eq.0) then
      if (namp1.ne.0) then
      open (2,form='formatted',file=fresp1)
      do 1 i=1,namp1
      read (2,*) afreq1(i),amp1(i)
1     continue
      close (2)
      endif
      if (namp2.ne.0) then
      open (2,form='formatted',file=fresp2)
      do 2 i=1,namp2
      read (2,*) afreq2(i),amp2(i)
2     continue
      close (2)
      endif
      iflag=1
      endif
c
      pi=3.141593
      w=2.*pi*f
      sa=const*ame*w*w/(1.+(f/f0)**2.) 
      if (f.eq.0.) gamma=0.
      if (f.ne.0.) then
      Q=Q0*f**eta
      gamma=w/(2.*Q*beta)
      endif
      anelas=exp(-gamma*r1)
      if (ikap.eq.0) highc=1./sqrt(1.+(f/fmax)**8.)
      if (ikap.eq.1) highc=exp(-pi*fmax*f)
      ga=1./r1
      if (igeom.ne.0) then
      if (r1.le.rg1) ga=r1**pow1
      if (r1.gt.rg1.and.r1.le.rg2) ga=(rg1**pow1)*(r1/rg1)**pow2
      if (r1.gt.rg2) ga=(rg1**pow1)*(rg2/rg1)**pow2*(r1/rg2)**pow3
      endif
c
      am1=1.
      am2=1.
      if (namp1.ne.0) am1=amplf(f,afreq1,amp1,namp1)
      if (namp2.ne.0) am2=amplf(f,afreq2,amp2,namp2)
      shape=sa*ga*highc*anelas*am1*am2
      return
      end
c
c
      subroutine window (i,nstart,nstop,ntaper,wind)
c     applies cosine tapered window
c     unit amplitude assumed
      wind=0.
      if (i.lt.nstart.or.i.gt.nstop) return
      wind=1.
      if(i.ge.nstart+ntaper.and.i.le.nstop-ntaper) return
      pi=3.141593
c     Value of wind goes from 0 to 1 from nstart to nstart+ntaper,
c     then from 1 to 0 from nstop-ntaper to nstop
      dum1=(nstop+nstart)/2.
      dum2=(nstop-nstart-ntaper)/2.
      wind=0.5*(1.-sin(pi*(abs(float(i)-dum1)-dum2)/float(ntaper)))
      return
      end
c
c
      subroutine wind2 (i,nstart,nstop,ntaper,tw,eps,eta,wind)
c     applies Saragoni and Hart (1974) window, with parameters
c     tw (dur), eps (fraction of dur to reach peak), and
c     eta (fraction of peak ampl. at tw)
c     The Saragoni and Hart window is applied between 
c     nstart + ntaper, and nstop - ntaper.  Outside these
c     bounds we do a quick cosine taper down to 0
c
      pi=3.141592654
      wind=0.
      if(i.lt.nstart.or.i.gt.nstop) return
      wind=1.
c
c     Apply Sargoni and Hart window
c
        b=-eps*alog(eta)/(1.+eps*(alog(eps)-1.))
        c=b/(eps*tw)
        a=(2.7182818/(eps*tw))**b
        if (i.ge.(nstart+ntaper).and.i.le.(nstop-ntaper)) then
          t=(tw/float(nstop-nstart-2*ntaper)) * 
     *                 (float(i-nstart-ntaper+1))
          wind=a*t**b*exp(-c*t)
          return
        else
c     Cos. taper goes from  0 to 1 from nstart to nstart+ntaper, from
c                           1 to 0 from nstop-ntaper to nstop.
c
        if (i.lt.(nstart+ntaper)) then
          t1=tw/float(nstop-nstart-2*ntaper)
          wf=a*t1**b*exp(-c*t)
          wind=abs(sin((float(i-nstart)/float(ntaper))*pi/2.))*wf
        else
          wf=a*tw**b*exp(-c*tw)
          wind=abs(sin((float(nstop-i)/float(ntaper))*pi/2.))*wf
        endif
        return
      endif
      end
c
c
      subroutine rscomp(a,dt,dur,damp,freq1,freq2,Nfreq,freq,psa)
c     calculates response spectrum from acceleration time history
      dimension Freq(*),Omega(100),psa(*),a(*)
      real pi,freq1,freq2,damp,Maxtim
c     damp is % critical damping. freq2 is maximum freq. to consider.
c
      pi=3.141593
      if (iabs(Nfreq).gt.100) then
         write (6,*)
         write (6,*) ' Nfreq cannot exceed 100'
         stop
      endif
c     Convert % damping to fraction
      beta=damp*0.01
c
c     Add 5 sec. to dur for shake-down
      maxtim=dur+5.
      n=ifix(dur/dt)
      ntime=ifix(maxtim/dt)
      do 100 j=n+1,ntime
        a(j)=0.
100   continue
c     Generate some logarithmically-spaced frequencies
       Freq(1)=freq1
       omega(1)=2.*pi*freq(1)
       if (Nfreq.ge.2) then
          frinc=alog(freq2/freq1)/(Nfreq-1)
          do 120 k=2,Nfreq
            freq(k)=freq1*exp((k-1)*frinc)
            omega(k)=2.*pi*freq(k)
120       continue
       endif
c        
c      Call new response spectrum routine for each desired freq
c
       do 200 k=1,nfreq
         call sdcomp(a,ntime,omega(k),beta,dt,sd)
         PSA(k)=sd*omega(k)*omega(k)
200    continue
       return
       end
c
c
      subroutine sdcomp(accg,na,omn,beta,dt,sd)
c This is a modified version of "Quake.For", written by
c Stavros A. Anagnostopoulos, Oct. 1986.  The formulation is that of
c Nigam and Jennings (BSSA, v. 59, 909-922, 1969).  This modification 
c eliminates the computation of the relative velocity and absolute 
c acceleration; it returns only the relative displacement.  
      dimension accg(*)
      omt=omn*dt
      d2=1-beta*beta
      d2=sqrt(d2)
      bom=beta*omn
      d3=2.*bom
      omd=omn*d2
      om2=omn*omn
      omdt=omd*dt
      c1=1./om2
      c2=2.*beta/(om2*omt)
      c3=c1+c2
      c4=1./(omn*omt)
      ss=sin(omdt)
      cc=cos(omdt)
      bomt=beta*omt
      ee=exp(-bomt)
      ss=ss*ee
      cc=cc*ee
      s1=ss/omd
      s2=s1*bom
      s3=s2+cc
      a11=s3
      a12=s1
      a21=-om2*s1
      a22=cc-s2
      s4=c4*(1.-s3)
      s5=s1*c4+c2
      b11=s3*c3-s5
      b12=-c2*s3+s5-c1
      b21=-s1+s4
      b22=-s4
      sd=0.
      n1=na-1
      y=0.
      ydot=0.
      do 1 i=1,n1
      y1=a11*y+a12*ydot+b11*accg(i)+b12*accg(i+1)
      ydot=a21*y+a22*ydot+b21*accg(i)+b22*accg(i+1)
c next two lines have been added for hardware portability
      if (y1.lt.1.e-30.and.y1.gt.0.) y1=0.
      if (ydot.lt.1.e-30.and.ydot.gt.0.) ydot=0.
      y=y1
      z=abs(y)
      z1=abs(ydot)
      if (z.gt.sd) sd=z
1     continue
      return
      end
c
c
	subroutine writa (facc,fpar,comm,trace,amax,lengts)
c	writes simulated acceleration time history into specified ascii file
        dimension trace(*)
	character facc*(*),fpar*(*),comm*(*)
	common /par/ leng,leng1,dt
	open (3,file=facc,status='unknown')
	write (3,40) comm,fpar
40	format (a70/'parameter file: ',a30)
	write (3,50) lengts
50	format (i5,' samples')
	write (3,30) dt
30	format ('dt: ',f6.3,' sec')
	write (3,'(a)') 'data format: f8.3,f11.4'
	write (3,10) amax
10	format ('maximum absolute acceleration:',f7.2) 
	write (3,'(a)') 'time (sec)  acceleration (cm/sec**2)'
	do 1 i=1,lengts
	time=(i-1)*dt
	write (3,20) time,trace(i)
1	continue
20	format (f8.3,f11.4)
	close (3)
	return
	end
c
c
	subroutine writp (fpar,comm,fi1,dip,h,alat1,alon1,alat2,alon2,
     *   rleng,wid,nl,nw,dl,dw,i0,j0,amag,sfact,stress,ntrig,slip)
c	writes modeling parameters to specified ascii file
        dimension slip(50,50)
	character typ*7,fpar*(*),fresp1*30,fresp2*30,comm*(*)
	common /par/ leng,leng1,dt
        common /sub/ ame,f0,rho,beta,ikap,fmax,Q0,eta,rise,rmin,rd1,
     *   rd2,durmin,b1,b2,b3,igeom,rg1,rg2,pow1,pow2,pow3,iwind,
     *   namp1,namp2
	common /fnames/ fresp1,fresp2
	open (3,file=fpar,status='unknown')
c write modeling parameters
	write (3,110) comm
110	format (a70/)
	write (3,10) fi1,dip,h
10	format ('fault strike, dip, depth to upper edge: ',3f6.1)
	write (3,20) rleng,wid
20	format ('fault length, width: ',2f7.1)
	write (3,170) nl,nw,dl,dw
170	format ('number of subfaults along strike and dip: ',2i4/
     *   'subfault length, width: ',2f6.1)
	write (3,30) i0,j0
30	format ('hypocenter at subfault: ',2i3)
	write (3,40) alat1,alon1,alat2,alon2
40	format ('lats, lons of origin and observation point: ' ,4f8.2)
	write (3,190) leng1,dt
190	format ('length of FFT: ',i6,' sampling interval: ',f6.3)
	write (3,50) beta,rho
50	format ('shear-wave velocity: ',f4.1,2x,' density: ',f4.1)
	write (3,180) sfact,stress
180	format ('radiation-strength factor: ',f6.2,
     *   ' stress parameter: ',f6.0)
	typ='fmax: '
	if (ikap.eq.1) typ='kapp: '
	write (3,60) typ,fmax,Q0,eta
60 	format (a7,f8.3,' Q0: ',f6.0,' eta: ',f6.2)
	write (3,140) rmin,rd1,rd2,durmin,b1,b2,b3
140	format ('duration model: ',' rmin, rd1, rd2: ',3f7.1,
     *   ' durmin: ',f5.2/'     b1, b2, b3: ',3f5.2)
	if (igeom.eq.0) write (3,160)
160	format ('geometric attenuation: 1/r')
	if (igeom.ne.0) write (3,150) rg1,rg2,pow1,pow2,pow3
150	format ('geom. att.: ',' rg1: ',f7.0,' rg2: ',f7.0,
     *   ' pow1: ',f5.1,' pow2: ',f5.1,' pow3: ',f5.1)
	if (iwind.eq.0) write (3,120)
	if (iwind.eq.1) write (3,130)
120	format ('window applied: tapered boxcar')
130 	format ('window applied: Saragoni-Hart')
	if (namp1.eq.0.and.namp2.eq.0) 
     *   write (3,'(a)') 'no local amplification'
	if (namp1.ne.0) write (3,200) fresp1
	if (namp2.ne.0) write (3,200) fresp2 
200	format ('amplification as in file ',a30)
	write (3,70) f0,rise
70	format ('subfault corner frequency: ',f7.2,
     *   ' subfault rise time: ',f6.1)
	write (3,80) amag,ntrig
80	format ('actual magnitude achieved: ',f4.1,
     *   ' total subsources summed: ',i5)
	write (3,90)
90	format ('actual slip distribution in meters: ')
	do 1 j=1,nw
	write (3,100) (slip(i,j),i=1,nl)
1	continue
100	format (50f5.1)
	close (3)
	return
	end
c
c
	subroutine writr (fpsa,fpar,comm,freq,psa,nfreq,ntrial,damp)
c	writes simulated average PSA spectrum at nfreq frequencies
c	into specified ascii file. First column is frequency in Hz, second 
c       column is PSA value in cm/s**2
        dimension freq(*),psa(*)
	character fpsa*(*),fpar*(*),comm*(*)
	open (3,file=fpsa,status='unknown')
	write (3,60) comm,fpar
60	format (a70/'parameter file: ',a30)
	write (3,40) ntrial
	write (3,50) int(damp)
	write (3,20) nfreq
	write (3,30) 
40	format ('average response spectrum over',i5,' trials')
50	format ('damping: ',i3,'%')
20	format (i4,' samples')
30	format ('data format: f10.2,f10.3'/
     *   ' freq (Hz)',t15,'PSA (cm/s**2)')
	write (3,10) (freq(i),psa(i),i=1,nfreq)
10	format (f10.2,f10.3)
	close (3)
	return
	end
c
c
	subroutine geom (alat1,alon1,alat2,alon2,epi,azi)
c	calculates distance and azimuth between two
c	points using their latitudes and longitudes
	re=6371.
	pi=3.141593
	alat=(alat1+alat2)/2.
	alat=pi/180.*alat
	dlat=alat2-alat1
	dlon=alon2-alon1
	r1=(2.*pi*re/360.)*(alat2-alat1)
	epi=(2.*pi*re/360.)*sqrt(dlat**2.+cos(alat)**2.*dlon**2.)
	azi=acos(r1/epi)
	if (dlon.le.0.) azi=2.*pi-azi
	azi=180./pi*azi
	return
	end
c
c
	real function amplf (f,afreq,amp,namp)
c	computes amplification value at frequency f by linear
c	interpolation between its adjoining values
	dimension amp(*),afreq(*)
	if (f.le.afreq(1)) then
	amplf=amp(1)
	return
	endif
	if (f.gt.afreq(namp)) then
	amplf=amp(namp)
	return
	endif
	do 1 i=2,namp
	if (f.le.afreq(i)) then
	d=(amp(i)-amp(i-1))/(afreq(i)-afreq(i-1))*(f-afreq(i-1))
	amplf=amp(i-1)+d
	return
	endif
1	continue
	end
